from __future__ import division
from builtins import zip
from builtins import range
from past.utils import old_div
__docformat__ = "restructuredtext en"

import mdp
from mdp import ClassifierNode, utils, numx, numx_rand, numx_linalg

# TODO: The GaussianClassifier and NearestMeanClassifier could be parallelized.


class SignumClassifier(ClassifierNode):
    """This classifier node classifies as ``1`` if the sum of the data points
    is positive and as ``-1`` if the data point is negative"""

    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('Float') +
                mdp.utils.get_dtypes('Integer'))

    @staticmethod
    def is_trainable():
        return False

    def _label(self, x):
        ret = [xi.sum() for xi in x]
        return numx.sign(ret)


class PerceptronClassifier(ClassifierNode):
    """A simple perceptron with input_dim input nodes."""
    
    def __init__(self, execute_method=None,
                 input_dim=None, output_dim=None, dtype=None):
        super(PerceptronClassifier, self).__init__(
                                                execute_method=execute_method,
                                                input_dim=input_dim,
                                                output_dim=output_dim,
                                                dtype=dtype)
        self.weights = []
        self.offset_weight = 0
        self.learning_rate = 0.1

    def _check_train_args(self, x, labels):
        if (isinstance(labels, (list, tuple, numx.ndarray)) and
            len(labels) != x.shape[0]):
            msg = ("The number of labels should be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)

        if (not isinstance(labels, (list, tuple, numx.ndarray))):
            labels = [labels]

        if (not numx.all([abs(x) == 1 for x in labels])):
            msg = "The labels must be either -1 or 1."
            raise mdp.TrainingException(msg)

    def _train(self, x, labels):
        """Update the internal structures according to the input data 'x'.

        x -- a matrix having different variables on different columns
             and observations on the rows.
        labels -- can be a list, tuple or array of labels (one for each data point)
              or a single label, in which case all input data is assigned to
              the same class.
        """

        # if weights are not yet initialised, initialise them
        if not len(self.weights):
            self.weights = numx.ones(self.input_dim)

        for xi, labeli in mdp.utils.izip_stretched(x, labels):
            new_weights = self.weights
            new_offset = self.offset_weight

            rate = self.learning_rate * (labeli - self._label(xi))
            for j in range(self.input_dim):
                new_weights[j] = self.weights[j] + rate * xi[j]

            # the offset corresponds to a node with input 1 all the time
            new_offset = self.offset_weight + rate * 1

            self.weights = new_weights
            self.offset_weight = new_offset

    def _label(self, x):
        """Returns an array with class labels from the perceptron.
        """
        return numx.sign(numx.dot(x, self.weights) + self.offset_weight)


class SimpleMarkovClassifier(ClassifierNode):
    """A simple version of a Markov classifier.
    It can be trained on a vector of tuples the label being the next element
    in the testing data.
    """
    def __init__(self, execute_method=None,
                 input_dim=None, output_dim=None, dtype=None):
        super(SimpleMarkovClassifier, self).__init__(
                                                execute_method=execute_method,
                                                input_dim=input_dim,
                                                output_dim=output_dim,
                                                dtype=dtype)
        self.ntotal_connections = 0

        self.features = {}
        self.labels = {}
        self.connections = {}
        
    def _get_supported_dtypes(self):
        """Return the list of dtypes supported by this node."""
        return (mdp.utils.get_dtypes('Float') +
                mdp.utils.get_dtypes('AllInteger') +
                mdp.utils.get_dtypes('Character'))

    def _check_train_args(self, x, labels):
        if (isinstance(labels, (list, tuple, numx.ndarray)) and
            len(labels) != x.shape[0]):
            msg = ("The number of labels should be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)

        if (not isinstance(labels, (list, tuple, numx.ndarray))):
            labels = [labels]

    def _train(self, x, labels):
        """Update the internal structures according to the input data 'x'.

        x -- a matrix having different variables on different columns
             and observations on the rows.
        labels -- can be a list, tuple or array of labels (one for each data point)
              or a single label, in which case all input data is assigned to
              the same class.
        """
        # if labels is a number, all x's belong to the same class
        for xi, labeli in mdp.utils.izip_stretched(x, labels):
            self._learn(xi, labeli)

    def _learn(self, feature, label):
        feature = tuple(feature)
        self.ntotal_connections += 1

        if label in self.labels:
            self.labels[label] += 1
        else:
            self.labels[label] = 1

        if feature in self.features:
            self.features[feature] += 1
        else:
            self.features[feature] = 1

        connection = (feature, label)
        if connection in self.connections:
            self.connections[connection] += 1
        else:
            self.connections[connection] = 1

    def _prob(self, features):
        return [self._prob_one(feature) for feature in features]

    def _prob_one(self, feature):
        feature = tuple(feature)
        probabilities = {}

        try:
            n_feature_connections = self.features[feature]
        except KeyError:
            n_feature_connections = 0
            # if n_feature_connections == 0, we get a division by zero
            # we could throw here, but maybe it's best to simply return
            # an empty dict object
            return {}

        for label in self.labels:
            conn = (feature, label)
            try:
                n_conn = self.connections[conn]
            except KeyError:
                n_conn = 0

            try:
                n_label_connections = self.labels[label]
            except KeyError:
                n_label_connections = 0

            p_feature_given_label = 1.0 * n_conn / n_label_connections
            p_label = 1.0 * n_label_connections / self.ntotal_connections
            p_feature = 1.0 * n_feature_connections / self.ntotal_connections
            prob = 1.0 * p_feature_given_label * p_label / p_feature
            probabilities[label] = prob
        return probabilities


class DiscreteHopfieldClassifier(ClassifierNode):
    """Node for simulating a simple discrete Hopfield model"""
    # TODO: It is unclear if this belongs to classifiers or is a general node
    # because label space is a subset of feature space
    def __init__(self, execute_method=None,
                 input_dim=None, output_dim=None, dtype='b'):
        super(DiscreteHopfieldClassifier, self).__init__(
                                            execute_method=execute_method,
                                            input_dim=input_dim,
                                            output_dim=output_dim,
                                            dtype=dtype)
        self._weight_matrix = 0 # assigning zero to ease addition
        self._num_patterns = 0
        self._shuffled_update = True

    def _get_supported_dtypes(self):
        return ['b']

    def _train(self, x):
        """Provide the hopfield net with the possible states.

        x -- a matrix having different variables on different columns
            and observations on rows.
        """
        for pattern in x:
            self._train_one(pattern)

    def _train_one(self, pattern):
        pattern = mdp.utils.bool_to_sign(pattern)
        weights = numx.outer(pattern, pattern)
        self._weight_matrix += old_div(weights, float(self.input_dim))
        self._num_patterns += 1

    @property
    def memory_size(self):
        """Returns the Hopfield net's memory size"""
        return self.input_dim

    @property
    def load_parameter(self):
        """Returns the load parameter of the Hopfield net.
        The quality of memory recall for a Hopfield net breaks down when the
        load parameter is larger than 0.14."""
        return old_div(self._num_patterns, float(self.input_dim))

    def _stop_training(self):
        # remove self-feedback
        # we could use numx.fill_diagonal, but thats numpy 1.4 only
        for i in range(self.input_dim):
            self._weight_matrix[i][i] = 0

    def _label(self, x, threshold = 0):
        """Retrieves patterns from the associative memory.
        """
        threshold = numx.zeros(self.input_dim) + threshold
        return numx.array([self._label_one(pattern, threshold) for pattern in x])

    def _label_one(self, pattern, threshold):
        pattern = mdp.utils.bool_to_sign(pattern)

        has_converged = False
        while not has_converged:
            has_converged = True
            iter_order = list(range(len(self._weight_matrix)))
            if self._shuffled_update:
                numx_rand.shuffle(iter_order)
            for row in iter_order:
                w_row = self._weight_matrix[row]

                thresh_row = threshold[row]
                new_pattern_row = numx.sign(numx.dot(w_row, pattern) - thresh_row)

                if new_pattern_row == 0:
                    # Following McKay, Neural Networks, we do nothing
                    # when the new pattern is zero
                    pass
                elif pattern[row] != new_pattern_row:
                    has_converged = False
                    pattern[row] = new_pattern_row
        return mdp.utils.sign_to_bool(pattern)

# TODO: Make it more efficient

class KMeansClassifier(ClassifierNode):
    """Employs K-Means Clustering for a given number of centroids."""
    def __init__(self, num_clusters, max_iter=10000, execute_method=None,
                 input_dim=None, output_dim=None, dtype=None):
        """
        :Arguments:
          num_clusters
            number of centroids to use = number of clusters
          max_iter
            if the algorithm does not reach convergence (for some
            numerical reason), stop after ``max_iter`` iterations
        """
        super(KMeansClassifier, self).__init__(execute_method=execute_method,
                                               input_dim=input_dim,
                                               output_dim=output_dim,
                                               dtype=dtype)
        self._num_clusters = num_clusters
        self.data = []
        self.tlen = 0
        self._centroids = None
        self.max_iter = max_iter

    def _train(self, x):
        # append all data
        # we could use a Cumulator class here
        self.tlen += x.shape[0]
        self.data.extend(x.ravel().tolist())

    def _stop_training(self):
        self.data = numx.array(self.data, dtype=self.dtype)
        self.data.shape = (self.tlen, self.input_dim)

        # choose initial centroids unless they are already given
        if not self._centroids:
            import random
            centr_idx = random.sample(range(self.tlen), self._num_clusters)
            #numx_rand.permutation(self.tlen)[:self._num_clusters]
            centroids = self.data[centr_idx]
        else:
            centroids = self._centroids

        for step in range(self.max_iter):
            # list of (sum_position, num_clusters)
            new_centroids = [(0., 0.)] * len(centroids)
            # cluster
            for x in self.data:
                idx = self._nearest_centroid_idx(x, centroids)
                # update position and count
                pos_count = (new_centroids[idx][0] + x,
                             new_centroids[idx][1] + 1.)
                new_centroids[idx] = pos_count

            # get new centroid position
            new_centroids = numx.array([old_div(c[0], c[1]) if c[1]>0. else centroids[idx]
                                        for idx, c in enumerate(new_centroids)])
            # check if we are stable
            if numx.all(new_centroids == centroids):
                self._centroids = centroids
                return
            centroids = new_centroids

    def _nearest_centroid_idx(self, data, centroids):
        dists = numx.array([numx.linalg.norm(data - c) for c in centroids])
        return dists.argmin()

    def _label(self, x):
        """For a set of feature vectors x, this classifier returns
        a list of centroids.
        """
        return [self._nearest_centroid_idx(xi, self._centroids) for xi in x]


class GaussianClassifier(ClassifierNode):
    """Perform a supervised Gaussian classification.

    Given a set of labelled data, the node fits a gaussian distribution
    to each class.
    """
    
    def __init__(self, execute_method=False,
                 input_dim=None, output_dim=None, dtype=None):
        super(GaussianClassifier, self).__init__(execute_method=execute_method,
                                                 input_dim=input_dim,
                                                 output_dim=output_dim,
                                                 dtype=dtype)
        self._cov_objs = {}  # only stored during training
        # this list contains the square root of the determinant of the
        # corresponding covariance matrix
        self._sqrt_def_covs = []
        # we are going to store the inverse of the covariance matrices
        # since only those are useful to compute the probabilities
        self.inv_covs = []
        self.means = []
        self.p = []  # number of observations
        self.labels = None

    @staticmethod
    def is_invertible():
        return False

    def _check_train_args(self, x, labels):
        if isinstance(labels, (list, tuple, numx.ndarray)) and (
            len(labels) != x.shape[0]):
            msg = ("The number of labels should be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)

    def _update_covs(self, x, lbl):
        if lbl not in self._cov_objs:
            self._cov_objs[lbl] = utils.CovarianceMatrix(dtype=self.dtype)
        self._cov_objs[lbl].update(x)

    def _train(self, x, labels):
        """
        :Arguments:
          x
              data
          labels
              Can be a list, tuple or array of labels (one for each data point)
              or a single label, in which case all input data is assigned to
              the same class.
        """
        # if labels is a number, all x's belong to the same class
        if isinstance(labels, (list, tuple, numx.ndarray)):
            labels_ = numx.asarray(labels)
            # get all classes from cl
            for lbl in set(labels_):
                x_lbl = numx.compress(labels_==lbl, x, axis=0)
                self._update_covs(x_lbl, lbl)
        else:
            self._update_covs(x, labels)

    def _stop_training(self):
        self.labels = list(self._cov_objs.keys())
        self.labels.sort()
        nitems = 0
        for lbl in self.labels:
            cov, mean, p = self._cov_objs[lbl].fix()
            nitems += p
            self._sqrt_def_covs.append(numx.sqrt(numx_linalg.det(cov)))
            if self._sqrt_def_covs[-1] == 0.0:
                err = ("The covariance matrix is singular for at least "
                       "one class.")
                raise mdp.NodeException(err)
            self.means.append(mean)
            self.p.append(p)
            self.inv_covs.append(utils.inv(cov))

        for i in range(len(self.p)):
            self.p[i] /= float(nitems)

        del self._cov_objs

    def _gaussian_prob(self, x, lbl_idx):
        """Return the probability of the data points x with respect to a
        gaussian.

        Input arguments:
        x -- Input data
        S -- Covariance matrix
        mn -- Mean
        """
        x = self._refcast(x)

        dim = self.input_dim
        sqrt_detS = self._sqrt_def_covs[lbl_idx]
        invS = self.inv_covs[lbl_idx]
        # subtract the mean
        x_mn = x - self.means[lbl_idx][numx.newaxis, :]
        # exponent
        exponent = -0.5 * (utils.mult(x_mn, invS)*x_mn).sum(axis=1)
        # constant
        constant = old_div((2.*numx.pi)**(old_div(-dim,2.)), sqrt_detS)
        # probability
        return constant * numx.exp(exponent)

    def class_probabilities(self, x):
        """Return the posterior probability of each class given the input."""
        self._pre_execution_checks(x)

        # compute the probability for each class
        tmp_prob = numx.zeros((x.shape[0], len(self.labels)),
                              dtype=self.dtype)
        for i in range(len(self.labels)):
            tmp_prob[:, i] = self._gaussian_prob(x, i)
            tmp_prob[:, i] *= self.p[i]

        # normalize to probability 1
        # (not necessary, but sometimes useful)
        tmp_tot = tmp_prob.sum(axis=1)
        tmp_tot = tmp_tot[:, numx.newaxis]
        return old_div(tmp_prob, tmp_tot)

    def _prob(self, x):
        """Return the posterior probability of each class given the input in a dict."""

        class_prob = self.class_probabilities(x)
        return [dict(list(zip(self.labels, prob))) for prob in class_prob]

    def _label(self, x):
        """Classify the input data using Maximum A-Posteriori."""

        class_prob = self.class_probabilities(x)
        winner = class_prob.argmax(axis=-1)
        return [self.labels[winner[i]] for i in range(len(winner))]
    
# TODO: Maybe extract some common elements form this class and
#    GaussianClassifier, like in _train.

class NearestMeanClassifier(ClassifierNode):
    """Nearest-Mean classifier."""
    
    def __init__(self, execute_method=None,
                 input_dim=None, output_dim=None, dtype=None):
        super(NearestMeanClassifier, self).__init__(
                                            execute_method=execute_method,
                                            input_dim=input_dim,
                                            output_dim=output_dim,
                                            dtype=dtype)
        self.label_means = {}  # not normalized during training
        self.n_label_samples = {}
        # initialized after training, used for vectorized execution:
        self.ordered_labels = []
        self.ordered_means = None  # will be array
        
    def _train(self, x, labels):
        """Update the mean information for the different classes.
        
        labels -- Can be a list, tuple or array of labels (one for each data
            point) or a single label, in which case all input data is assigned
            to the same class (computationally this is more efficient).
        """
        if isinstance(labels, (list, tuple, numx.ndarray)):
            labels = numx.asarray(labels)
            for label in set(labels):
                x_label = numx.compress(labels==label, x, axis=0)
                self._update_mean(x_label, label)
        else:
            self._update_mean(x, labels)
            
    def _update_mean(self, x, label):
        """Update the mean with data for a single label."""
        if label not in self.label_means:
            self.label_means[label] = numx.zeros(self.input_dim)
            self.n_label_samples[label] = 0
        # TODO: use smarter summing to avoid rounding errors
        self.label_means[label] += numx.sum(x, axis=0)
        self.n_label_samples[label] += len(x)
        
    def _check_train_args(self, x, labels):
        if isinstance(labels, (list, tuple, numx.ndarray)) and (
            len(labels) != x.shape[0]):
            msg = ("The number of labels should be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)
        
    def _stop_training(self):
        """Calculate the class means."""
        ordered_means = [] 
        for label in self.label_means:
            self.label_means[label] /= self.n_label_samples[label]
            self.ordered_labels.append(label)
            ordered_means.append(self.label_means[label])
        self.ordered_means = numx.vstack(ordered_means)
            
    def _label(self, x):
        """Classify the data based on minimal distance to mean."""
        n_labels = len(self.ordered_labels)
        differences = x[:,:,numx.newaxis].repeat(n_labels, 2). \
                        swapaxes(1,2) - self.ordered_means
        square_distances = (differences**2).sum(2)
        label_indices = square_distances.argmin(1)
        labels = [self.ordered_labels[i] for i in label_indices]
        return labels
    
    
class KNNClassifier(ClassifierNode):
    """K-Nearest-Neighbour Classifier."""
    
    def __init__(self, k=1, execute_method=None,
                 input_dim=None, output_dim=None, dtype=None):
        """Initialize classifier.
        
        k -- Number of closest sample points that are taken into account.
        """
        super(KNNClassifier, self).__init__(execute_method=execute_method,
                                            input_dim=input_dim,
                                            output_dim=output_dim,
                                            dtype=dtype)
        self.k = k
        self._label_samples = {}  # temporary variable during training
        self.n_samples = None
        # initialized after training:
        self.samples = None  # 2d array with all samples
        self.sample_label_indices = None  # 1d array for label indices
        self.ordered_labels = []
        
    def _train(self, x, labels):
        """Add the sampel points to the classes.
        
        labels -- Can be a list, tuple or array of labels (one for each data
            point) or a single label, in which case all input data is assigned
            to the same class (computationally this is more efficient).
        """
        if isinstance(labels, (list, tuple, numx.ndarray)):
            labels = numx.asarray(labels)
            for label in set(labels):
                x_label = numx.compress(labels==label, x, axis=0)
                self._add_samples(x_label, label)
        else:
            self._add_samples(x, labels)
    
    def _add_samples(self, x, label):
        """Store x set for later neirest-neighbour calculation."""
        if label not in self._label_samples:
            self._label_samples[label] = []
        self._label_samples[label].append(x)
        
    def _check_train_args(self, x, labels):
        if isinstance(labels, (list, tuple, numx.ndarray)) and (
            len(labels) != x.shape[0]):
            msg = ("The number of labels should be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)
        
    def _stop_training(self):
        """Organize the sample data."""
        ordered_samples = []
        for label in self._label_samples:
            ordered_samples.append(
                            numx.concatenate(self._label_samples[label]))
            self.ordered_labels.append(label)
        del self._label_samples
        self.samples = numx.concatenate(ordered_samples)
        self.n_samples = len(self.samples)
        self.sample_label_indices = numx.concatenate(
                                [numx.ones(len(ordered_samples[i]),
                                           dtype="int32") * i
                                 for i in range(len(self.ordered_labels))])

    def _label(self, x):
        """Label the data by comparison with the reference points."""
        square_distances = (x*x).sum(1)[:, numx.newaxis] \
                      + (self.samples*self.samples).sum(1)
        square_distances -= 2 * numx.dot(x, self.samples.T)
        min_inds = square_distances.argsort()
        win_inds = [numx.bincount(self.sample_label_indices[indices[0:self.k]]).
                    argmax(0) for indices in min_inds]
        labels = [self.ordered_labels[i] for i in win_inds]
        return labels



"""
Module to trace and document the training and execution of a BiFlow.

This module supports (Bi)HiNet structures. Monkey patching is used to
inject the tracing code into the Flow.

InspectionHTMLTracer is the main class. It uses TraceDecorationVisitor to add
the tracing decoration to the flow and TraceHTMLConverter to create HTML view
of the flow state (which in turn uses TraceHTMLVisitor for the flow
representation).

Note that this module does not combine the trace views into a slideshow, this
is done in the seperate slideshow module.
"""

# TODO: wrap inner methods (e.g. _train) to document effective arguments?

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object

import os
import pickle as pickle
import fnmatch
import copy
import traceback

import mdp
n = mdp.numx
import mdp.hinet as hinet

from bimdp import BiNode
from bimdp import BiFlow
from bimdp.hinet import BiFlowNode, CloneBiLayer

from bimdp.hinet import BiHiNetHTMLVisitor
from .utils import robust_pickle

CLICKABLE_NODE_ID = "clickable_node_%d"
# standard css filename for the complete CSS:
STANDARD_CSS_FILENAME = "mdp.css"

NODE_TRACE_METHOD_NAMES = ["execute", "train", "stop_training"]
BINODE_TRACE_METHOD_NAMES = []  # methods that are only traced in binodes
TRACING_WRAP_FLAG = "_insp_is_wrapped_for_tracing_"
ORIGINAL_METHOD_PREFIX = "_insp_original_"


class TraceDebugException(Exception):
    """Exception for return the information when debug is True."""

    def __init__(self, result):
        """Store the information necessary to finish the tracing.

        result -- The result that would otherwise be returned by the method.
        """
        super(TraceDebugException, self).__init__()
        self.result = result


class InspectionHTMLTracer(object):
    """Class for inspecting a single pass through a provided flow.

    This class is based on a visitor that decorates the flow elements with
    tracing wrappers. It also provides a callback function for the tracers
    and stores everything else needed for the inspection.

    This class is already specialized for creating HTML slides in the callback
    function.

    Note that a flow decorated for tracing is not compatible with pickling
    or parallel training and execution. Normally the decorated flow is
    only used in trace_training or trace_execution anyway.
    """

    def __init__(self, html_converter=None, css_filename=STANDARD_CSS_FILENAME):
        """Prepare for tracing and create the HTML translator.

        html_converter -- TraceHTMLConverter instance, with a convert_flow
            method to create the flow visualization for each slide.
        css_filename -- CSS file used for all the slides
            (default 'inspect.css').
        """
        if html_converter is None:
            self._html_converter = TraceHTMLConverter()
        else:
            self._html_converter = html_converter
        self._css_filename = css_filename
        self._tracing_decorator = TraceDecorationVisitor(
                            decorator=self._standard_tracer_decorate,
                            undecorator=self._standard_tracer_undecorate)
        self._trace_path = None  # path for the current trace
        self._trace_name = None  # name for the current trace
        self._flow = None  # needed for the callback HTML translation
        # step counter used in the callback, is reset automatically
        self._slide_index = None
        self._slide_filenames = None
        self._section_ids = None  # can be used during execution
        self._slide_node_ids = None  # active node for each slide index

    def _reset(self):
        """Reset the internal variables for a new tracing.

        Should be called before 'train', 'stop_training' or 'execute' is called
        on the flow.
        """
        self._slide_index = 0
        self._slide_filenames = []
        self._section_ids = []
        self._slide_node_ids = []
        self._html_converter.reset()

    def trace_training(self, path, flow, x, msg=None, stop_msg=None,
                       trace_name="training", debug=False, **kwargs):
        """Trace a single training phase and the stop_training.

        Return a tuple containing a list of the training slide filenames, the
        training node ids and the same for stop_training.

        path -- Path were the inspection files will be stored.
        trace_name -- Name prefix for this inspection (default is training).
        **kwargs -- Additional arguments for flow.train can be specified
            as keyword arguments.
        """
        self._reset()
        self._trace_path = path
        # train and stop filenames must be different
        self._trace_name = trace_name + "_t"
        self._flow = flow
        self._tracing_decorator.decorate_flow(flow)
        biflownode = BiFlowNode(BiFlow(flow.flow))
        try:
            biflownode.train(x=x, msg=msg, **kwargs)
            # reset is important for the following stop_training
            biflownode.bi_reset()
        # Note: this also catches legacy string exceptions (which are still
        #    used in numpy, e.g. np.core.multiarray.error)
        except:
            if debug:
                # insert the error slide and encapsulate the exception
                traceback.print_exc()
                self._write_error_frame()
                result = (self._slide_filenames, self._slide_node_ids,
                          None, None)
                raise TraceDebugException(result=result)
            else:
                raise
        train_filenames = self._slide_filenames
        train_node_ids = self._slide_node_ids
        self._reset()
        self._trace_name = trace_name + "_s"
        try:
            biflownode.stop_training(stop_msg)
        except:
            if debug:
                # insert the error slide and encapsulate the exception
                traceback.print_exc()
                self._write_error_frame()
                result = (train_filenames, train_node_ids,
                          self._slide_filenames, self._slide_node_ids)
                raise TraceDebugException(result=result)
            else:
                raise
        stop_filenames = self._slide_filenames
        stop_node_ids = self._slide_node_ids
        # restore undecorated flow
        self._tracing_decorator.decorate_flow(flow, undecorate_mode=True)
        return train_filenames, train_node_ids, stop_filenames, stop_node_ids

    def trace_execution(self, path, trace_name, flow, x, msg=None, target=None,
                        debug=False, **kwargs):
        """Trace a single execution.

        The return value is a tuple containing a list of the slide filenames,
        the node ids, the section_ids for a slideshow with sections
        (or None if no section_ids were used) and the execution output value.

        path -- Path were the inspection files will be stored.
        trace_name -- Name prefix for this inspection.
        **kwargs -- Additional arguments for flow.execute can be specified
            as keyword arguments.
        """
        self._reset()
        self._trace_path = path
        self._trace_name = trace_name
        self._flow = flow
        self._tracing_decorator.decorate_flow(flow)
        if (not (isinstance(flow, BiFlow) or isinstance(flow, BiNode)) and
            (msg is not None)):
            # a msg would be interpreted as nodenr by a Flow, so check this
            err = "A msg was given for a normal Flow (need BiFlow)."
            raise Exception(err)
        try:
            if msg or target:
                result = self._flow.execute(x, msg, target, **kwargs)
            # this case also works for mdp.Flow
            else:
                result = self._flow.execute(x, **kwargs)
        # Note: this also catches legacy string exceptions (which are still
        #    used in numpy, e.g. np.core.multiarray.error)
        except:
            if debug:
                # insert the error slide and encapsulate the exception
                traceback.print_exc()
                self._write_error_frame()
                if not self._section_ids:
                    self._section_ids = None
                result = (self._slide_filenames, self._slide_node_ids,
                          self._section_ids)
                raise TraceDebugException(result=result)
            else:
                raise
        self._tracing_decorator.decorate_flow(flow, undecorate_mode=True)
        if not self._section_ids:
            self._section_ids = None
        else:
            if len(self._section_ids) != len(self._slide_filenames):
                err = ("Mismatch between number of section_ids and number of "
                       "slides.")
                raise Exception(err)
        return (self._slide_filenames, self._slide_node_ids,
                self._section_ids, result)

    def _tracer_callback(self, node, method_name, method_result, method_args,
                         method_kwargs):
        """This method is called by the tracers.

        The calling tracer also provides this method with the needed state
        information and the method arguments.

        node -- The node from which the callback was initiated.
        method_name -- Name of the method from which the callback was initiated.
        result -- Return value of the method.
        args, kwargs -- The arguments of the method call.
        """
        ## write visualization to html_file
        try:
            html_file = self._begin_HTML_frame()
            section_id, node_id = self._html_converter.write_html(
                                            path=self._trace_path,
                                            html_file=html_file,
                                            flow=self._flow,
                                            node=node,
                                            method_name=method_name,
                                            method_result=method_result,
                                            method_args=method_args,
                                            method_kwargs=method_kwargs)
            self._slide_index += 1
            if section_id is not None:
                self._section_ids.append(section_id)
            self._slide_node_ids.append(node_id)
        finally:
            self._end_HTML_frame(html_file)

    ## HTML decoration ##

    def _begin_HTML_frame(self):
        """Return the HTML file for a trace frame including the header.

        The file should then be finished via _end_HTML_frame.
        """
        path = self._trace_path
        filename = self._trace_name + "_%d.html" % self._slide_index
        self._slide_filenames.append(filename)
        html_file = open(os.path.join(path, filename), "w")
        html_file = hinet.NewlineWriteFile(html_file)
        html_file.write('<html>\n<head>\n<title>Inspection Slide</title>')
        if self._css_filename:
            html_file.write('<style type="text/css" media="screen">')
            html_file.write('@import url("%s");' % self._css_filename)
            html_file.write('</style>\n</head>\n<body>')
        return html_file

    def _end_HTML_frame(self, html_file):
        """Complete and close the HTML file for a trace frame.

        The method should always be used after _begin_HTML_frame.
        """
        html_file.write('</body>\n</html>')

    def _write_error_frame(self):
        with self._begin_HTML_frame() as html_file:
            html_file.write('<div class="error">')
            html_file.write('<h3>Encountered Exception</h3>')
            traceback_html = traceback.format_exc().replace('\n', '<br>')
#        get HTML traceback, didn't work due to legacy stuff
#        TODO: retry this in the future
#        import StringIO as stringio
#        import cgitb
#        import mdp
#        exception_type, exception, tb = sys.exc_info()
#        # Problem: only the text of the original exception is stored in
#        #     mdp.FlowExceptionCR, and the text is not even correctpy displayed.
##        if exception_type is mdp.FlowExceptionCR:
##            exception.args = tuple()
##            exception.message = None
#        buffer = stringio.StringIO()
#        handler = cgitb.Hook(file=buffer)
#        handler.handle((exception_type, exception, tb))
#        traceback_html = buffer.getvalue()
            html_file.write(traceback_html)
            html_file.write('</div>')
            self._end_HTML_frame(html_file)

    ## monkey patching tracing decorator wrapper methods ##

    def _standard_tracer_decorate(self, node):
        """Adds a tracer wrapper to the node via monkey patching."""
        # add a marker to show that this node is wrapped
        setattr(node, TRACING_WRAP_FLAG, True)
        trace_method_names = list(NODE_TRACE_METHOD_NAMES)
        if isinstance(node, BiNode):
            trace_method_names += BINODE_TRACE_METHOD_NAMES
        for method_name in trace_method_names:
            new_method_name = ORIGINAL_METHOD_PREFIX + method_name
            # create a reference to the original method
            setattr(node, new_method_name, getattr(node, method_name))
            # use nested scopes  lexical closure to get proper wrapper
            def get_wrapper(_method_name, _inspector):
                _new_method_name = ORIGINAL_METHOD_PREFIX + method_name
                def wrapper(self, *args, **kwargs):
                    args_copy = copy.deepcopy(args)
                    kwargs_copy = copy.deepcopy(kwargs)
                    result = getattr(self, _new_method_name)(*args, **kwargs)
                    _inspector._tracer_callback(self, _method_name, result,
                                                args_copy, kwargs_copy)
                    return result
                return wrapper
            # hide the original method in this instance behind the wrapper
            setattr(node, method_name,
                    get_wrapper(method_name, self).__get__(node))
        # modify getstate to enable pickling (get rid of the instance methods)
        def wrapped_getstate(self):
            result = self.__dict__.copy()
            if not hasattr(node, TRACING_WRAP_FLAG):
                return result
            del result[TRACING_WRAP_FLAG]
            # delete all instance methods
            trace_method_names = list(NODE_TRACE_METHOD_NAMES)
            if isinstance(self, BiNode):
                trace_method_names += BINODE_TRACE_METHOD_NAMES
            for method_name in trace_method_names:
                del result[method_name]
                old_method_name = ORIGINAL_METHOD_PREFIX + method_name
                del result[old_method_name]
            del result["__getstate__"]
            return result
        node.__getstate__ = wrapped_getstate.__get__(node)

    def _standard_tracer_undecorate(self, node):
        """Remove a tracer wrapper from the node."""
        if not hasattr(node, TRACING_WRAP_FLAG):
            return
        delattr(node, TRACING_WRAP_FLAG)
        trace_method_names = list(NODE_TRACE_METHOD_NAMES)
        if isinstance(node, BiNode):
            trace_method_names += BINODE_TRACE_METHOD_NAMES
        for method_name in trace_method_names:
            # delete the wrapped method in the instance to unhide the original
            delattr(node, method_name)
            # delete the no longer used reference to the original method
            old_method_name = ORIGINAL_METHOD_PREFIX + method_name
            delattr(node, old_method_name)
        # restore normal getstate
        delattr(node, "__getstate__")


class TraceDecorationVisitor(object):
    """Class to add tracing wrappers to nodes in a flow."""

    def __init__(self, decorator, undecorator):
        """Initialize.

        decorator -- Callable decorator that wraps node methods.
        undecorator -- Callable decorator that removes the wrapper from a
            method.
        """
        self._decorator = decorator
        self._undecorator = undecorator
        # note that _visit_clonelayer uses the undecorate mode
        self._undecorate_mode = None

    def decorate_flow(self, flow, undecorate_mode=False):
        """Adds or removes wrappers from the nodes in the given flow."""
        self._undecorate_mode = undecorate_mode
        for node in flow:
            self._visit_node(node)

    def _visit_node(self, node):
        if hasattr(node, "flow"):
            self._visit_flownode(node)
        elif isinstance(node, mdp.hinet.CloneLayer):
            self._visit_clonelayer(node)
        elif isinstance(node, mdp.hinet.Layer):
            self._visit_layer(node)
        else:
            self._visit_standard_node(node)

    def _visit_standard_node(self, node):
        """Wrap the node."""
        if not self._undecorate_mode:
            self._decorator(node)
        else:
            self._undecorator(node)

    def _visit_flownode(self, flownode):
        for node in flownode.flow:
            self._visit_node(node)

    def _visit_layer(self, layer):
        for node in layer:
            self._visit_node(node)

    def _visit_clonelayer(self, clonelayer):
        # TODO: enable the use of a shallow copy to save memory,
        #    but this requires to implement __copy__ in Node etc. for recursive
        #    shallow copying
        if self._undecorate_mode:
            if isinstance(clonelayer, CloneBiLayer):
                # check that clonelayer is actually decorated
                if not hasattr(clonelayer, "_original_set_use_copies"):
                    return
                del clonelayer._set_use_copies
                del clonelayer._original_set_use_copies
                del clonelayer.__getstate__
                self._visit_node(clonelayer.nodes[0])
                if not clonelayer.use_copies:
                    clonelayer.nodes = ((clonelayer.node,) *
                                        len(clonelayer.nodes))
            else:
                self._visit_node(clonelayer.nodes[0])
                clonelayer.nodes = (clonelayer.node,) * len(clonelayer.nodes)
            # undecoration is complete
            return
        ## decorate clonelayer
        if ((not isinstance(clonelayer, CloneBiLayer)) or
            (not clonelayer.use_copies)):
            # use a decorated deep copy for the first node
            clonelayer.node = clonelayer.nodes[0].copy()
            clonelayer.nodes = (clonelayer.node,) + clonelayer.nodes[1:]
        # only decorate the first node
        self._visit_node(clonelayer.nodes[0])
        if isinstance(clonelayer, CloneBiLayer):
            # add a wrapper to _set_use_copies,
            # otherwise all nodes in layer would get decorated
            clonelayer._original_set_use_copies = clonelayer._set_use_copies
            flow_decorator = self
            def wrapped_use_copies(self, use_copies):
                # undecorate internal nodes to allow copy operation
                flow_decorator._undecorate_mode = True
                flow_decorator._visit_node(clonelayer.nodes[0])
                flow_decorator._undecorate_mode = False
                if use_copies and not self.use_copies:
                    # switch to node copies, no problem
                    clonelayer._original_set_use_copies(use_copies)
                elif not use_copies and self.use_copies:
                    # switch to a single node instance
                    # but use a (decorated) deep copy for first node
                    clonelayer._original_set_use_copies(use_copies)
                    clonelayer.node = clonelayer.nodes[0].copy()
                    clonelayer.nodes = ((clonelayer.node,) +
                                        clonelayer.nodes[1:])
                flow_decorator._visit_node(clonelayer.nodes[0])
            clonelayer._set_use_copies = wrapped_use_copies.__get__(clonelayer)
            # modify getstate to enable pickling
            # (get rid of the instance methods)
            def wrapped_getstate(self):
                result = self.__dict__.copy()
                # delete instance methods
                del result["_original_set_use_copies"]
                del result["_set_use_copies"]
                del result["__getstate__"]
                return result
            clonelayer.__getstate__ = wrapped_getstate.__get__(clonelayer)


_INSPECTION_CSS_FILENAME = "trace.css"

def inspection_css():
    """Return the CSS for the inspection slides."""
    css_filename = os.path.join(os.path.split(__file__)[0],
                                _INSPECTION_CSS_FILENAME)
    with open(css_filename, 'r') as css_file:
        css = css_file.read()
    return BiHiNetHTMLVisitor.hinet_css() + css


class TraceHTMLVisitor(BiHiNetHTMLVisitor):
    """Special BiHiNetHTMLVisitor to take into account runtime info.

    It highlights the currently active node.
    """

    def __init__(self, html_file, show_size=False):
        super(TraceHTMLVisitor, self).__init__(html_file,
                                               show_size=show_size)
        self._current_node = None
        self._node_id_index = None
        # this is the HTML node id, not the Node attribute
        self._current_node_id = None

    def convert_flow(self, flow, current_node=None):
        self._current_node = current_node
        self._node_id_index = 0
        self._current_node_id = None
        super(TraceHTMLVisitor, self).convert_flow(flow)

    def _open_node_env(self, node, type_id="node"):
        """Open the HTML environment for the node internals.

        This special version highlights the nodes involved in the trace.

        node -- The node itself.
        type_id -- The id string as used in the CSS.
        """
        f = self._file
        html_line = '<table class="'
        trace_class = None
        if node is self._current_node:
            trace_class = "current_node"
        elif type_id == "node" and node._train_phase_started:
            trace_class = "training_node"
        if trace_class:
            html_line += ' %s' % trace_class
        html_line += ' %s' % type_id
        # assign id only to nodes which trigger a slide creation,
        # i.e. only if the node can become active
        if hasattr(node, TRACING_WRAP_FLAG):
            node_id = CLICKABLE_NODE_ID % self._node_id_index
            if node is self._current_node:
                self._current_node_id = node_id
            self._node_id_index += 1
            html_line +=  ' clickable" id="%s">' % node_id
        else:
            html_line += '">'
        f.write(html_line)
        self._write_node_header(node, type_id)


class TraceHTMLConverter(object):
    """Class to visualize the state of a BiFlow during execution or training.

    The single snapshot is a beefed up version of the standard HTML view.
    Capturing the data to make this possible is not the responsibility of this
    class.
    """

    def __init__(self, flow_html_converter=None):
        """Initialize the internal variables."""
        if flow_html_converter is None:
            self.flow_html_converter = TraceHTMLVisitor(html_file=None)
        else:
            self.flow_html_converter= flow_html_converter
        self._html_file = None

    def reset(self):
        """Reset internal variables for a new trace.

        It is called (by TraceHTMLInspector) before calling 'train',
        'stop_training' or 'execute' on the flow.

        This method can be overridden by derived that need to keep track of the
        training or execution phase.
        """
        pass

    @staticmethod
    def _array_pretty_html(ar):
        """Return a nice HTML representation of the given numpy array."""
        ar_str = 'array with shape %s<br>\n' % str(ar.shape)
        # TODO: use np.savetxt instead?
        ar_str += (str(ar).replace(' [', '<br>\n[').
                    replace(']\n ...', ']<br>\n...'))
        return ar_str

    @classmethod
    def _dict_pretty_html(cls, dic):
        """Return a nice HTML representation of the given numpy array."""
        # TODO: use an stringio buffer for efficency
        # put array keys last, because arrays are typically rather large
        keys = [key for key, value in list(dic.items())
                if not isinstance(value, n.ndarray)]
        keys.sort()
        ar_keys = [key for key, value in list(dic.items())
                   if isinstance(value, n.ndarray)]
        ar_keys.sort()
        keys += ar_keys
        dic_strs = []
        for key in keys:
            value = dic[key]
            dic_str = '<span class="keyword">' + repr(key) + '</span>: '
            if isinstance(value, str):
                dic_str += repr(value)
            elif isinstance(value, n.ndarray):
                dic_str += cls._array_pretty_html(value)
            else:
                dic_str += str(value)
            dic_strs.append(dic_str)
        return '{' + ',<br>\n'.join(dic_strs) + '}'

    def write_html(self, path, html_file, flow, node, method_name,
                    method_result, method_args, method_kwargs):
        """Write the HTML translation of the flow into the provided file.

        Return value is the section_id and the HTML/CSS id of the active node.
        The section id is ignored during training.

        path -- Path of the slide (e.h. to store additional images).
        html_file -- File of current slide, where the translation is written.
        flow -- The overall flow.
        node -- The node that was called last.
        method_name -- The method that was called on the last node.
        method_result -- The result from the last call.
        method_args -- args that were given to the method
        method_kwargs -- kwargs that were given to the method
        """
        self._html_file = hinet.NewlineWriteFile(html_file)
        f = self._html_file
        ## create table, left side for the flow, right side for data
        f.write('<br><br>')
        f.write('<table><tr><td id="inspect_biflow_td">')
        f.write("<h3>flow state</h3>")
        self.flow_html_converter._file = f
        self.flow_html_converter.convert_flow(flow, node)
        # now the argument / result part of the table
        f.write('</td><td id="inspect_result_td">')
        section_id = self._write_data_html(
                               path=path, html_file=html_file, flow=flow,
                               node=node, method_name=method_name,
                               method_result=method_result,
                               method_args=method_args,
                               method_kwargs=method_kwargs)
        f.write('</table>')
        f.write('</td></tr>\n</table>')
        self._html_file = None
        return section_id, self.flow_html_converter._current_node_id

    def _write_data_html(self, path, html_file, flow, node, method_name,
                         method_result, method_args, method_kwargs):
        """Write the data part (right side of the slide).

        Return value can be a section_id or None. The section_id is ignored
        during training (since the slideshow sections are used for the
        training phases).

        This method can be overriden for custom visualisations. Usually this
        original method should still be called via super.

        path -- Path of the slide (e.h. to store additional images).
        html_file -- File of current slide, where the translation is written.
        flow -- The overall flow.
        node -- The node that was called last.
        method_name -- The method that was called on the last node.
        method_result -- The result from the last call.
        method_args -- args that were given to the method
        method_kwargs -- kwargs that were given to the method
        """
        f = self._html_file
        f.write('<h3>%s arguments</h3>' % method_name)
        f.write('<table class="inspect_io_data">')
        if method_name == "stop_training":
            # first argument is not x,
            # if no arguments were given method_args == (None,)
            if method_args == (None,):
                f.write('<tr><td><pre>None</pre></tr></td>')
        else:
            # deal and remove x part of arguments
            x = method_args[0]
            if isinstance(x, n.ndarray):
                f.write('<tr><td><pre>x = </pre></td>' +
                        '<td>' + self._array_pretty_html(x) + '</td></tr>')
            else:
                f.write('<tr><td><pre>x = </pre></td><td>' + str(x) +
                        '</td></tr>')
        # remaining arg is message
        method_args = method_args[1:]
        if method_args and method_args[0] is not None:
            f.write('<tr><td><pre>msg = </pre></td><td>' +
                    self._dict_pretty_html(method_args[0]) + '</td></tr>')
        # normally the kwargs should be empty
        for arg_key in method_kwargs:
            f.write('<tr><td><pre>' + arg_key + ' = </pre></td><td>' +
                    str(method_kwargs[arg_key]) + '</td></tr>')
        f.write('</table>')
        ## print results
        f.write("<h3>%s result</h3>" % method_name)
        f.write('<table class="inspect_io_data">')
        if method_result is None:
            f.write('<tr><td><pre>None</pre></tr></td>')
        elif isinstance(method_result, n.ndarray):
            f.write('<tr><td><pre>x = </pre></td><td>' +
                    self._array_pretty_html(method_result) + '</td></tr>')
        elif isinstance(method_result, tuple):
            f.write('<tr><td><pre>x = </pre></td><td>')
            if isinstance(method_result[0], n.ndarray):
                f.write(self._array_pretty_html(method_result[0]) +
                        '</td></tr>')
            else:
                f.write(str(method_result[0]) + '</td></tr>')
            # second value is msg
            f.write('<tr><td><pre>msg = </pre></td><td>')
            if isinstance(method_result[1], dict):
                f.write(self._dict_pretty_html(method_result[1]) +
                        '</td></tr>')
            else:
                f.write(str(method_result[1]) + '</td></tr>')
            # last value is target
            if len(method_result) > 2:
                f.write('<tr><td><pre>target = </pre></td><td>' +
                        str(method_result[2]) + '</td></tr>')
        else:
            f.write('<tr><td><pre>unknown result type: </pre></td><td>' +
                    str(method_result) + '</td></tr>')


## Functions to capture pickled biflow snapshots during training. ##

PICKLE_EXT = ".pckl"
PICKLE_PROTO = -1
SNAPSHOT_FILENAME = "snapshot"

def prepare_training_inspection(flow, path):
    """Use hook in the BiFlow to store a snapshot in each training phase.

    path -- Path were the snapshots are stored.

    This is done by wrapping the _stop_training_hook method of biflow.
    Some attributes are added to the biflow which store all information needed
    for the pickling (like filename). To enable pickling we use the
    __getstate__ slot, since some attributes cannot be pickled.
    """
    # add attributes to biflow which are used in wrapper_method
    flow._snapshot_counter_ = 0
    flow._snapshot_path_ = path
    flow._snapshot_name_ = SNAPSHOT_FILENAME
    flow._snapshot_instance_methods_ = []
    ### wrap _stop_training_hook to store biflow snapshots ###
    def pickle_wrap_method(_flow, _method_name):
        new_method_name = ORIGINAL_METHOD_PREFIX + _method_name
        def wrapper(self, *args, **kwargs):
            result = getattr(self, new_method_name)(*args, **kwargs)
            # pickle biflow
            filename = (self._snapshot_name_ + "_%d" % self._snapshot_counter_ +
                        PICKLE_EXT)
            robust_pickle(self._snapshot_path_, filename, self)
            self._snapshot_counter_ += 1
            return result
        # create a reference to the original method
        setattr(_flow, new_method_name, getattr(_flow, _method_name))
        # hide the original method in this instance behind the wrapper
        setattr(_flow, _method_name, wrapper.__get__(_flow))
        _flow._snapshot_instance_methods_.append(_method_name)
        _flow._snapshot_instance_methods_.append(new_method_name)
    pickle_wrap_method(flow, "_stop_training_hook")
    ### wrap __getstate__ to enable pickling ###
    # note that in the pickled flow no trace of the wrapping remains
    def wrapped_biflow_getstate(self):
        result = self.__dict__.copy()
        # delete all instancemethods
        for method_name in self._snapshot_instance_methods_:
            del result[method_name]
        # delete the special attributes which were inserted by the wrapper
        # (not really necessary)
        del result["_snapshot_counter_"]
        del result["_snapshot_path_"]
        del result["_snapshot_name_"]
        del result["_snapshot_instance_methods_"]
        # remove data attributes (generators cannot be pickled)
        # pop with default value also works when key is not present in dict
        result.pop("_train_data_iterables", None)
        result.pop("_train_data_iterator", None)
        result.pop("_train_msg_iterables", None)
        result.pop("_train_msg_iterator", None)
        result.pop("_stop_messages", None)
        result.pop("_exec_data_iterator", None)
        result.pop("_exec_msg_iterator", None)
        result.pop("_exec_target_iterator", None)
        return result
    flow.__getstate__ = wrapped_biflow_getstate.__get__(flow)
    flow._snapshot_instance_methods_.append("__getstate__")

def remove_inspection_residues(flow):
    """Remove all the changes made by prepare_training_inspection."""
    try:
        for method_name in flow._snapshot_instance_methods_:
            delattr(flow, method_name)
        del flow._snapshot_counter_
        del flow._snapshot_path_
        del flow._snapshot_name_
        del flow._snapshot_instance_methods_
    except:
        # probably the hooks were already removed, so do nothing
        pass

def _trace_biflow_training(snapshot_path, inspection_path,
                           x_samples, msg_samples=None, stop_messages=None,
                           tracer=None,
                           debug=False, show_size=False, verbose=True,
                           **kwargs):
    """Load flow snapshots and perform the inspection with the given data.

    The return value consists of the slide filenames, the slide node ids,
    and an index table (index of last slide of section indexed by node,
    phase, train and stop). If no snapshots were found the return value is
    None.

    snapshot_path -- Path were the flow training snapshots are stored.
    inspection_path -- Path were the slides are stored.
    css_filename -- Filename of the CSS file for the slides.
    x_samples, msg_samples -- Lists with the input data for the training trace.
    stop_messages -- The stop msg for the training trace.
    tracer -- Instance of InspectionHTMLTracer, can be None for
        default class.
    debug -- If True (default is False) then any exception will be
        caught and the gathered data up to that point is returned in the
        normal way. This is useful for bimdp debugging.
    show_size -- Show the approximate memory footprint of all nodes.
    verbose -- If True (default value) a status message is printed for each
        loaded snapshot.
    **kwargs -- Additional arguments for flow.train can be specified
        as keyword arguments.
    """
    if not tracer:
        tracer = InspectionHTMLTracer()
        tracer._html_converter.flow_html_converter.show_size = show_size
    i_train_node = 0  # index of the training node
    i_snapshot = 0 # snapshot counter
    index_table = [[]]  # last slide indexed by [node, phase, train 0 or stop 1]
    slide_filenames = []
    slide_node_ids = []
    try:
        # search for the snapshot files
        for file_path, dirs, files in os.walk(os.path.abspath(snapshot_path)):
            dirs.sort()
            files = fnmatch.filter(files, SNAPSHOT_FILENAME + "*" + PICKLE_EXT)
            files.sort()
            for filename in files:
                filename = os.path.join(file_path, filename)
                # load the flow snapshot
                biflow = None  # free memory
                with open(filename, "rb") as pickle_file:
                    biflow = pickle.load(pickle_file)
                # determine which node is training and set the indices
                for node in biflow[i_train_node:]:
                    if node.get_remaining_train_phase() > 0:
                        break
                    else:
                        i_train_node += 1
                        index_table.append([])
                # inspect the training
                x = x_samples[i_train_node]
                if msg_samples:
                    msg = msg_samples[i_train_node]
                else:
                    msg = None
                if stop_messages:
                    stop_msg = stop_messages[i_train_node]
                else:
                    stop_msg = None
                trace_name = "%d_%d" % (i_snapshot, i_train_node)
                train_files, train_ids, stop_files, stop_ids = \
                    tracer.trace_training(trace_name=trace_name,
                                          path=inspection_path,
                                          flow=biflow,
                                          x=x, msg=msg, stop_msg=stop_msg,
                                          debug=debug,
                                          **kwargs)
                slide_filenames += train_files
                train_index = len(slide_filenames) - 1
                slide_filenames += stop_files
                stop_index = len(slide_filenames) - 1
                index_table[i_train_node].append((train_index, stop_index))
                slide_node_ids += train_ids
                slide_node_ids += stop_ids
                if verbose:
                    print("got traces for snapshot %d" % (i_snapshot + 1))
                i_snapshot += 1
    except TraceDebugException as debug_exception:
        train_files, train_ids, stop_files, stop_ids = debug_exception.result
        slide_filenames += train_files
        train_index = len(slide_filenames) - 1
        if stop_files:
            slide_filenames += stop_files
        stop_index = len(slide_filenames) - 1
        index_table[i_train_node].append((train_index, stop_index))
        slide_node_ids += train_ids
        if stop_ids:
            slide_node_ids += stop_ids
        debug_exception.result = (slide_filenames, slide_node_ids, index_table)
        raise
    return slide_filenames, slide_node_ids, index_table



from __future__ import division
from builtins import range
from past.utils import old_div
from ._tools import *

def verify_ICANode(icanode, rand_func = uniform, vars = 3, N=8000,
                 prec = 3):
    dim = (N,vars)
    mat,mix,inp = get_random_mix(rand_func=rand_func,mat_dim=dim)
    icanode.train(inp)
    act_mat = icanode.execute(inp)
    cov = utils.cov2(old_div((mat-mean(mat,axis=0)),std(mat,axis=0)), act_mat)
    maxima = numx.amax(abs(cov), axis=0)
    assert_array_almost_equal(maxima,numx.ones(vars),prec)

def verify_ICANodeMatrices(icanode, rand_func = uniform, vars = 3, N=8000):
    dim = (N,vars)
    mat,mix,inp = get_random_mix(rand_func=rand_func,
                                       mat_dim=dim, avg = 0)
    icanode.train(inp)
    # test projection matrix
    act_mat = icanode.execute(inp)
    T = icanode.get_projmatrix()
    exp_mat = mult(inp, T)
    assert_array_almost_equal(act_mat,exp_mat,6)
    # test reconstruction matrix
    out = act_mat.copy()
    act_mat = icanode.inverse(out)
    B = icanode.get_recmatrix()
    exp_mat = mult(out, B)
    assert_array_almost_equal(act_mat,exp_mat,6)

def rand_with_timestruct(size=None):
    T, N = size
    # do something special only if T!=N, otherwise
    # we were asked to generate a mixing matrix
    if T == N:
        return uniform(size=size)
    # create independent sources
    src = uniform((T,N))*2-1
    fsrc = numx_fft.rfft(src,axis=0)
    # enforce different speeds
    for i in range(N):
        fsrc[(i+1)*(T//20):,i] = 0.
    src = numx_fft.irfft(fsrc,axis=0)
    return src

def test_CuBICANode_batch():
    ica = mdp.nodes.CuBICANode(limit=10**(-decimal))
    ica2 = ica.copy()
    verify_ICANode(ica)
    verify_ICANodeMatrices(ica2)

def test_CuBICANode_telescope():
    ica = mdp.nodes.CuBICANode(limit=10**(-decimal), telescope=1)
    ica2 = ica.copy()
    verify_ICANode(ica)
    verify_ICANodeMatrices(ica2)


def test_TDSEPNode():
    ica = mdp.nodes.TDSEPNode(lags=20, limit=1e-10)
    ica2 = ica.copy()
    verify_ICANode(ica, rand_func=rand_with_timestruct, vars=2, N=2**14, prec=2)
    verify_ICANodeMatrices(ica2, rand_func=rand_with_timestruct, vars=2, N=2**14)



import mdp
from mdp import PreserveDimNode, numx, VariadicCumulator
import operator


class ClassifierNode(PreserveDimNode):
    """A ClassifierNode can be used for classification tasks that should not
    interfere with the normal execution flow. A reason for that is that the
    labels used for classification do not form a vector space, and so they don't
    make much sense in a flow.
    """
    
    def __init__(self, execute_method=None,
                 input_dim=None, output_dim=None, dtype=None):
        """Initialize classifier.
        
        execute_method -- Set to string value 'label', 'rank', or 'prob' to
            force the corresponding classification method being used instead
            of the standard identity execution (which is used when
            execute_method has the default value None). This can be used when
            the node is last in a flow, the return value from Flow.execute
            will then consist of the classification results.
        """
        self.execute_method = execute_method
        super(ClassifierNode, self).__init__(input_dim=input_dim,
                                             output_dim=output_dim,
                                             dtype=dtype)

    ### Methods to be implemented by the subclasses

    def _label(self, x, *args, **kargs):
        raise NotImplementedError

    def _prob(self, x, *args, **kargs):
        raise NotImplementedError

    ### User interface to the overwritten methods

    def label(self, x, *args, **kwargs):
        """Returns an array with best class labels.

        By default, subclasses should overwrite _label to implement
        their label. The docstring of the '_label' method
        overwrites this docstring.
        """
        self._pre_execution_checks(x)
        return self._label(self._refcast(x), *args, **kwargs)

    def prob(self, x, *args, **kwargs):
        """Returns the probability for each datapoint and label
        (e.g., [{1:0.1, 2:0.0, 3:0.9}, {1:1.0, 2:0.0, 3:0.0}, ...])

        By default, subclasses should overwrite _prob to implement
        their prob. The docstring of the '_prob' method
        overwrites this docstring.
        """
        self._pre_execution_checks(x)
        return self._prob(self._refcast(x), *args, **kwargs)

    def rank(self, x, threshold=None):
        """Returns ordered list with all labels ordered according to prob(x)
        (e.g., [[3 1 2], [2 1 3], ...]).

        The optional threshold parameter is used to exclude labels having equal
        or less probability. E.g. threshold=0 excludes all labels with zero
        probability.
        """
        all_ranking = []
        prob = self.prob(x)
        for p in prob:
            if threshold is None:
                ranking = list(p.items())
            else:
                ranking = ((k, v) for k, v in list(p.items()) if v > threshold)
            result = [k for k, v in
                      sorted(ranking, key=operator.itemgetter(1), reverse=True)]
            all_ranking.append(result)
        return all_ranking
    
    def _execute(self, x):
        if not self.execute_method:
            return x
        elif self.execute_method == "label":
            return self.label(x)
        elif self.execute_method == "rank":
            return self.rank(x)
        elif self.execute_method == "prob":
            return self.prob(x)

# XXX are the _train and _stop_training functions necessary anymore?
class ClassifierCumulator(VariadicCumulator('data', 'labels'), ClassifierNode):
    """A ClassifierCumulator is a Node whose training phase simply collects
    all input data and labels. In this way it is possible to easily implement
    batch-mode learning.

    The data is accessible in the attribute 'self.data' after
    the beginning of the '_stop_training' phase. 'self.tlen' contains
    the number of data points collected.
    'self.labels' contains the assigned label to each data point.
    """

    def __init__(self, input_dim=None, output_dim=None, dtype=None):
        super(ClassifierCumulator, self).__init__(input_dim=input_dim,
                                                  output_dim=output_dim,
                                                  dtype=dtype)

    def _check_train_args(self, x, labels):
        super(ClassifierCumulator, self)._check_train_args(x, labels)
        if (isinstance(labels, (list, tuple, numx.ndarray)) and
            len(labels) != x.shape[0]):
            msg = ("The number of labels must be equal to the number of "
                   "datapoints (%d != %d)" % (len(labels), x.shape[0]))
            raise mdp.TrainingException(msg)

    def _train(self, x, labels):
        """Cumulate all input data in a one dimensional list."""
        self.tlen += x.shape[0]
        self.data.extend(x.ravel().tolist())

        # if labels is a number, all x's belong to the same class
        if isinstance(labels, (list, tuple, numx.ndarray)):
            pass
        else:
            labels = [labels] * x.shape[0]

        self.labels.extend(labels.ravel().tolist())

    def _stop_training(self, *args, **kwargs):
        """Transform the data and labels lists to array objects and reshape them."""
        self.data = numx.array(self.data, dtype=self.dtype)
        self.data.shape = (self.tlen, self.input_dim)
        self.labels = numx.array(self.labels)
        self.labels.shape = (self.tlen)




"""
Some helper functions and classes for inspection.
"""
from future import standard_library
standard_library.install_aliases()
from builtins import next
from builtins import object

import os
import pickle as pickle


def robust_pickle(path, filename, obj):
    """Robust pickle function, creates path if it does not exist."""
    filename = os.path.join(path, filename)
    try:
        picke_file = open(filename, "wb")
    except IOError as inst:
        error_code = inst.args[0]
        if error_code == 2:  # path does not exist
            os.makedirs(path)
            picke_file = open(filename, "wb")
        else:
            raise
    try:
        pickle.dump(obj, picke_file, -1)
    finally:
        picke_file.close()

def robust_write_file(path, filename, content):
    """Create a file with the given content and return the filename.

    If the provided path does not exist it will be created.
    If the file already exists it will be overwritten.
    """
    try:
        new_file = open(os.path.join(path, filename), "w")
    except IOError as inst:
        error_code = inst.args[0]
        if error_code == 2:  # path does not exist
            os.makedirs(path)
            new_file = open(os.path.join(path, filename), "w")
        else:
            raise
    new_file.write(content)
    return filename

def first_iterable_elem(iterable):
    """Helper function to get the first element of an iterator or iterable.

    The return value is a tuple of the first element and the iterable.
    If the iterable is actually an iterator then a decorator is used to wrap
    it and extract the first element in a non-consuming way.
    """
    if iter(iterable) is iterable:
        # iterable is actually iterator, have to wrap it
        peek_iter = PeekIterator(iterable)
        first_elem = peek_iter.peek()
        return first_elem, peek_iter
    else:
        first_elem = next(iter(iterable))
        return first_elem, iterable


class PeekIterator(object):
    """Look-ahead iterator decorator."""

    def __init__(self, iterator):
        self.iterator = iterator
        # we simplicity we do not use collections.deque
        self.cache = []

    def peek(self):
        """Return the next element in the iterator without consuming it.

        So the returned elements will still be returned by next in the normal
        order. If the iterator has no next element then the StopIterator
        exception is passed.
        """
        next_elem = next(self)
        # TODO: use a dequeue for better efficiency
        self.cache = [next_elem] + self.cache
        return next_elem

    def __next__(self):
        if self.cache:
            return self.cache.pop()
        else:
            return next(self.iterator)

    def __iter__(self):
        return self



from builtins import range

import mdp
import mdp.hinet as hinet
n = mdp.numx

from bimdp import BiNode, BiNodeException


class CloneBiLayerException(BiNodeException):
    """CloneBiLayer specific exception."""
    pass


class CloneBiLayer(BiNode, hinet.CloneLayer):
    """BiMDP version of CloneLayer.

    Since all the nodes in the layer are identical, it is guaranteed that the
    target identities match. The outgoing data on the other hand is not checked.
    So if the notes return different kinds of results the overall result is very
    unpredictable.

    The incoming data is split into len(self.nodes) parts, so the actual chunk
    size does not matter as long as it is compatible with this scheme.
    This also means that this class can deal with incoming data from a
    BiSwitchboard that is being send down.
    Arrays in the message are split up if they can be evenlty split into
    len(self.nodes) parts along the second axis, otherwise they are put
    into each node message. Arrays in the outgoing message are joined along
    the second axis (unless they are the same unsplit array), so if an array
    is accidently split no harm should be done (there is only some overhead).

    Note that a msg is always passed to the internal nodes, even if the Layer
    itself was targeted. Additional target resolution can then happen in the
    internal node (e.g. like it is done in the standard BiFlowNode).
    
    Both incomming and outgoing messages are automatically checked for the
    the use_copies msg key.
    """

    def __init__(self, node, n_nodes=1, use_copies=False,
                 node_id=None, dtype=None):
        """Initialize the internal variables.

        node -- Node which makes up the layer.
        n_nodes -- Number of times the node is repeated in this layer.
        use_copies -- Determines if a single instance or copies of the node
            are used.
        """
        super(CloneBiLayer, self).__init__(node_id=node_id, node=node,
                                           n_nodes=n_nodes, dtype=dtype)
        # (self.node is None) is used as flag for self.use_copies
        self.use_copies = use_copies

    use_copies = property(fget=lambda self: self._get_use_copies(),
                          fset=lambda self, flag: self._set_use_copies(flag))

    def _get_use_copies(self):
        """Return the use_copies flag."""
        return self.node is None

    def _set_use_copies(self, use_copies):
        """Switch internally between using a single node instance or copies.

        In a normal CloneLayer a single node instance is used to represent all
        the horizontally aligned nodes. But in a BiMDP where the nodes store
        temporary data this may not work.
        Via this method one can therefore create copies of the single node
        instance.

        This method can also be triggered by the use_copies msg key.
        """
        if use_copies and (self.node is not None):
            # switch to node copies
            self.nodes = [self.node.copy() for _ in range(len(self.nodes))]
            self.node = None  # disable single node while copies are used
        elif (not use_copies) and (self.node is None):
            # switch to a single node instance
            if self.is_training():
                err = ("Calling switch_to_instance during training will "
                       "probably result in lost training data.")
                raise CloneBiLayerException(err)
            elif self.is_bi_training():
                err = ("Calling switch_to_instance during bi_learning will "
                       "probably result in lost learning data.")
                raise CloneBiLayerException(err)
            self.node = self.nodes[0]
            self.nodes = [self.node] * len(self.nodes)

    def _get_method(self, method_name, default_method, target):
        """Return the default method and the unaltered target.

        This method overrides the standard BiNode _get_method to delegate the
        method selection to the internal nodes.
        """
        return default_method, target

    ## standard node methods ##

    def _check_input(self, x):
        """Input check is disabled.

        It will be checked by the targeted internal node.
        """
        pass

    def _execute(self, x, msg=None):
        """Process the data through the internal nodes."""
        if msg is not None:
            self._extract_message_copy_flag(msg)
        y_results = []
        msg_results = []
        target = None
        node_msgs = self._get_split_messages(msg)
        if x is not None:
            # use the dimension of x, because this also works for inverse
            node_dim = x.shape[1] // len(self.nodes)
        else:
            node_dim = None
        for i_node, node in enumerate(self.nodes):
            if node_dim:
                node_x = x[:, node_dim*i_node : node_dim*(i_node+1)]
            else:
                node_x = None
            node_msg = node_msgs[i_node]
            if node_msg:
                node_result = node.execute(node_x, node_msg)
            else:
                node_result = node.execute(node_x)
            ## store result
            if not isinstance(node_result, tuple):
                y_results.append(node_result)
            else:
                y_results.append(node_result[0])
                msg_results.append(node_result[1])
                if len(node_result) == 3:
                    target = node_result[2]
        ## combine message results
        msg = self._get_combined_message(msg_results)
        if (not y_results) or (y_results[-1] is None):
            y = None
        else:
            y = n.hstack(y_results)
        # check outgoing message for use_copies key
        if msg is not None:
            self._extract_message_copy_flag(msg)
        ## return result
        if target is not None:
            return (y, msg, target)
        elif msg:
            return (y, msg)
        else:
            return y

    def _train(self, x, msg=None):
        """Perform single training step by training the internal nodes."""
        ## this code is mostly identical to the execute code,
        ## currently the only difference is that train is called
        if msg is not None:
            self._extract_message_copy_flag(msg)
        y_results = []
        msg_results = []
        target = None
        node_msgs = self._get_split_messages(msg)
        if x is not None:
            # use the dimension of x, because this also works for inverse
            node_dim = x.shape[1] // len(self.nodes)
        else:
            node_dim = None
        for i_node, node in enumerate(self.nodes):
            if node_dim:
                node_x = x[:, node_dim*i_node : node_dim*(i_node+1)]
            else:
                node_x = None
            node_msg = node_msgs[i_node]
            if node_msg:
                node_result = node.train(node_x, node_msg)
            else:
                node_result = node.train(node_x)
            ## store result
            if not isinstance(node_result, tuple):
                y_results.append(node_result)
            else:
                y_results.append(node_result[0])
                msg_results.append(node_result[1])
                if len(node_result) == 3:
                    target = node_result[2]
        ## combine message results
        msg = self._get_combined_message(msg_results)
        if (not y_results) or (y_results[-1] is None):
            y = None
        else:
            y = n.hstack(y_results)
        # check outgoing message for use_copies key
        if msg is not None:
            self._extract_message_copy_flag(msg)
        ## return result
        if target is not None:
            return (y, msg, target)
        elif msg:
            return (y, msg)
        else:
            return y

    def _stop_training(self, msg=None):
        """Call stop_training on the internal nodes.

        The outgoing result message is also searched for a use_copies key,
        which is then applied if found.
        """
        if msg is not None:
            self._extract_message_copy_flag(msg)
        target = None
        if self.use_copies:
            ## have to call stop_training for each node
            y_results = []
            msg_results = []
            node_msgs = self._get_split_messages(msg)
            for i_node, node in enumerate(self.nodes):
                node_msg = node_msgs[i_node]
                if node_msg:
                    node_result = node.stop_training(node_msg)
                else:
                    node_result = node.stop_training()
                ## store result
                if not isinstance(node_result, tuple):
                    y_results.append(node_result)
                else:
                    y_results.append(node_result[0])
                    msg_results.append(node_result[1])
                    if len(node_result) == 3:
                        target = node_result[2]
            ## combine message results
            msg = self._get_combined_message(msg_results)
            if (not y_results) or (y_results[-1] is None):
                y = None
            else:
                y = n.hstack(y_results)
        else:  
            ## simple case of a single instance
            node_result = self.node.stop_training(msg)
            if not isinstance(node_result, tuple):
                return node_result
            elif len(node_result) == 2:
                y, msg = node_result
            else:
                y, msg, target = node_result
        # check outgoing message for use_copies key
        if msg is not None:
            self._extract_message_copy_flag(msg)
        # return result
        if target is not None:
            return (y, msg, target)
        elif msg:
            return (y, msg)
        else:
            return y

    ## BiNode methods ##

    def _bi_reset(self):
        """Call bi_reset on all the inner nodes."""
        if self.use_copies:
            for node in self.nodes:
                node.bi_reset()
        else:
            # note: reaching this code probably means that copies should be used
            self.node.bi_reset()

    def _request_node_id(self, node_id):
        """Return an internal node if it matches the provided node id.

        If the node_id matches that of the layer itself, then self is returned.
        """
        if self.node_id == node_id:
            return self
        if not self.use_copies:
            return self.node._request_node_id(node_id)
        else:
            # return the first find, but call _request_node_id on all copies
            # otherwise BiFlowNode._last_id_request would get confused
            first_found_node = None
            for node in self.nodes:
                found_node = node._request_node_id(node_id)
                if (not first_found_node) and found_node:
                    first_found_node = found_node
            return first_found_node

    ## Helper methods for message handling ##

    def _extract_message_copy_flag(self, msg):
        """Look for the the possible copy flag and modify the msg if needed.

        If the copy flag is found the Node is switched accordingly.
        """
        msg_id_keys = self._get_msg_id_keys(msg)
        copy_flag = self._extract_message_key("use_copies", msg, msg_id_keys)
        if copy_flag is not None:
            self.use_copies = copy_flag

    def _get_split_messages(self, msg):
        """Return messages for the individual nodes."""
        if not msg:
            return [None] * len(self.nodes)
        msgs = [dict() for _ in range(len(self.nodes))]
        n_nodes = len(self.nodes)
        for (key, value) in list(msg.items()):
            if (isinstance(value, n.ndarray) and
                # check if the array can be split up
                len(value.shape) >= 2 and not value.shape[1] % n_nodes):
                # split the data along the second index
                split_values = n.hsplit(value, n_nodes)
                for i, split_value in enumerate(split_values):
                    msgs[i][key] = split_value
            else:
                for node_msg in msgs:
                    # Note: the value is not copied, just referenced
                    node_msg[key] = value
        return msgs

    def _get_combined_message(self, msgs):
        """Return the combined message.

        Only keys from the last entry in msgs are used. Only when the value
        is an array are all the msg values combined.
        """
        if (not msgs) or (msgs[-1] is None):
            return None
        if len(msgs) == 1:
            return msgs[0]
        msg = dict()
        for (key, one_value) in list(msgs[-1].items()):
            other_value = msgs[0][key]
            if (isinstance(one_value, n.ndarray) and
                # check if the array was originally split up
                (len(one_value.shape) >= 2 and one_value is not other_value)):
                msg[key] = n.hstack([node_msg[key] for node_msg in msgs])
            else:
                # pick the msg value of the last node
                msg[key] = msgs[-1][key]
        return msg



"""
Extension for building switchboards in a 2d hierarchical network.
"""
from builtins import range

# TODO: add unittests and maybe mention it in the tutorial

# TODO: maybe integrate all this into the original switchboard classes?

import mdp
from mdp.hinet import (
    ChannelSwitchboard, Rectangular2dSwitchboard, DoubleRect2dSwitchboard,
    DoubleRhomb2dSwitchboard
)

def get_2d_image_switchboard(image_size_xy):
    """Return a Rectangular2dSwitchboard representing an image.

    This can then be used as the prev_switchboard.
    """
    return Rectangular2dSwitchboard(in_channels_xy=image_size_xy,
                                    field_channels_xy=1,
                                    field_spacing_xy=1)


class FactoryExtensionChannelSwitchboard(mdp.ExtensionNode,
                                         ChannelSwitchboard):
    """Extension node for the assembly of channel switchboards.

    data attributes:
    free_parameters -- List of parameters that do not depend on the previous
        layer. Note that there might still be restrictions imposed by the
        switchboard.
        By convention parameters that end with '_xy' can either be a single
        int or a 2-tuple (sequence) of ints.
    compatible_pre_switchboards -- List of compatible base classes for
        prev_switchboard.
    """

    extension_name = "switchboard_factory"

    free_parameters = []
    compatible_pre_switchboards = [ChannelSwitchboard]

    @classmethod
    def create_switchboard(cls, free_params, prev_switchboard,
                           prev_output_dim):
        """Return a new instance of this switchboard.

        free_params -- Parameters as specified by free_parameters.
        prev_switchboard -- Instance of the previous switchboard.
        prev_output_dim -- Output dimension of the previous layer.

        This template method checks the compatibility of the prev_switchboard
        and sanitizes '_xy' in free_params.
        """
        compatible = False
        for base_class in cls.compatible_pre_switchboards:
            if isinstance(prev_switchboard, base_class):
                compatible = True
        if not compatible:
            err = ("The prev_switchboard class '%s'" %
                        prev_switchboard.__class__.__name__ +
                   " is not compatible with this switchboard class" +
                   " '%s'." % cls.__name__)
            raise mdp.hinet.SwitchboardException(err)
        kwargs = cls._get_switchboard_kwargs(free_params, prev_switchboard,
                                             prev_output_dim)
        return cls(**kwargs)

    @staticmethod
    def _get_switchboard_kwargs(free_params, prev_switchboard,
                                prev_output_dim):
        """Return the kwargs for the cls '__init__' method.

        Reference implementation, merges input into one single channel.
        Override this method for other switchboard classes.
        """
        in_channel_dim = prev_output_dim // prev_switchboard.output_channels
        return {"input_dim": prev_output_dim,
                "connections": list(range(prev_output_dim)),
                "out_channel_dim": prev_output_dim,
                "in_channel_dim": in_channel_dim}


class FactoryRectangular2dSwitchboard(FactoryExtensionChannelSwitchboard,
                                      Rectangular2dSwitchboard):

    free_parameters = ["field_channels_xy", "field_spacing_xy", "ignore_cover"]
    compatible_pre_switchboards = [Rectangular2dSwitchboard,
                                   DoubleRhomb2dSwitchboard]

    @staticmethod
    def _get_switchboard_kwargs(free_params, prev_switchboard,
                                prev_output_dim):
        in_channel_dim = (prev_output_dim // prev_switchboard.output_channels)
        if not "ignore_cover" in free_params:
            free_params["ignore_cover"] = True
        return {"in_channels_xy": prev_switchboard.out_channels_xy,
                "field_channels_xy": free_params["field_channels_xy"],
                "field_spacing_xy": free_params["field_spacing_xy"],
                "in_channel_dim": in_channel_dim,
                "ignore_cover": free_params["ignore_cover"]}


class FactoryDoubleRect2dSwitchboard(FactoryExtensionChannelSwitchboard,
                                     DoubleRect2dSwitchboard):

    free_parameters = ["field_channels_xy", "ignore_cover"]
    compatible_pre_switchboards = [Rectangular2dSwitchboard,
                                   DoubleRhomb2dSwitchboard]

    @staticmethod
    def _get_switchboard_kwargs(free_params, prev_switchboard,
                                prev_output_dim):
        in_channel_dim = (prev_output_dim // prev_switchboard.output_channels)
        if not "ignore_cover" in free_params:
            free_params["ignore_cover"] = True
        return {"in_channels_xy": prev_switchboard.out_channels_xy,
                "field_channels_xy": free_params["field_channels_xy"],
                "in_channel_dim": in_channel_dim,
                "ignore_cover": free_params["ignore_cover"]}


class FactoryDoubleRhomb2dSwitchboard(FactoryExtensionChannelSwitchboard,
                                      DoubleRhomb2dSwitchboard):

    free_parameters = ["field_size"]
    compatible_pre_switchboards =  [DoubleRect2dSwitchboard]

    @staticmethod
    def _get_switchboard_kwargs(free_params, prev_switchboard,
                                prev_output_dim):
        in_channel_dim = (prev_output_dim // prev_switchboard.output_channels)
        return {"long_out_channels_xy": prev_switchboard.long_out_channels_xy,
                "diag_field_channels": free_params["field_size"],
                "in_channel_dim": in_channel_dim}
